Optimal - objective value 46.43125659
      0 TotalDiscountedCost(BB,2016)                                                 46.431257                       0
      1 RateOfDemand(BB,x,electricity,2016)                                             2.1101                       0
      2 NewCapacity(BB,gas_import,2016)                                               2.342422                       0
      3 AccumulatedNewCapacity(BB,gas_import,2016)                                    2.342422                       0
      6 TotalCapacityAnnual(BB,gas_import,2016)                                       2.342422                       0
      7 TotalCapacityAnnual(BB,gas_plant,2016)                                          3.1101                       0
      8 RateOfActivity(BB,x,gas_import,1,2016)                                        2.342422                       0
      9 RateOfTotalActivity(BB,gas_import,x,2016)                                     2.342422                       0
     10 RateOfActivity(BB,x,gas_plant,1,2016)                                           2.1101                       0
     11 RateOfTotalActivity(BB,gas_plant,x,2016)                                        2.1101                       0
     12 RateOfProductionByTechnologyByMode(BB,x,gas_import,1,natural_gas,2016)        2.342422                       0
     13 RateOfProductionByTechnologyByMode(BB,x,gas_plant,1,electricity,2016)           2.1101                       0
     14 RateOfProductionByTechnology(BB,x,gas_import,natural_gas,2016)                2.342422                       0
     15 RateOfProductionByTechnology(BB,x,gas_plant,electricity,2016)                   2.1101                       0
     16 RateOfProduction(BB,x,natural_gas,2016)                                       2.342422                       0
     17 RateOfProduction(BB,x,electricity,2016)                                         2.1101                       0
     18 RateOfUseByTechnologyByMode(BB,x,gas_plant,1,natural_gas,2016)                2.342422                       0
     19 RateOfUseByTechnology(BB,x,gas_plant,natural_gas,2016)                        2.342422                       0
     21 RateOfUse(BB,x,natural_gas,2016)                                              2.342422                       0
     22 Production(BB,x,natural_gas,2016)                                             2.342422                       0
     23 Production(BB,x,electricity,2016)                                               2.1101                       0
     24 Use(BB,x,natural_gas,2016)                                                    2.342422                       0
     25 Demand(BB,x,electricity,2016)                                                   2.1101                       0
     28 ProductionAnnual(BB,natural_gas,2016)                                         2.342422                       0
     29 ProductionAnnual(BB,electricity,2016)                                           2.1101                       0
     30 UseAnnual(BB,natural_gas,2016)                                                2.342422                       0
     32 ProductionByTechnology(BB,x,gas_import,natural_gas,2016)                      2.342422                       0
     33 ProductionByTechnology(BB,x,gas_plant,electricity,2016)                         2.1101                       0
     34 UseByTechnology(BB,x,gas_plant,natural_gas,2016)                              2.342422                       0
     35 TotalAnnualTechnologyActivityByMode(BB,gas_import,1,2016)                     2.342422                       0
     36 TotalAnnualTechnologyActivityByMode(BB,gas_plant,1,2016)                        2.1101                       0
     37 ModelPeriodCostByRegion(BB)                                                  46.431257                       0
     38 CapitalInvestment(BB,gas_import,2016)                                     2.342422e-05                       0
     40 DiscountedCapitalInvestment(BB,gas_import,2016)                           2.342422e-05                       0
     46 AnnualVariableOperatingCost(BB,gas_plant,2016)                               19.244534                       0
     48 AnnualFixedOperatingCost(BB,gas_plant,2016)                                  28.333322                       0
     51 OperatingCost(BB,gas_plant,2016)                                             47.577856                       0
     53 DiscountedOperatingCost(BB,gas_plant,2016)                                   46.431233            1.110223e-16
     54 TotalDiscountedCostByTechnology(BB,gas_import,2016)                       2.342422e-05                       0
     56 TotalDiscountedCostByTechnology(BB,gas_plant,2016)                           46.431233                       0
     58 TotalTechnologyAnnualActivity(BB,gas_import,2016)                             2.342422                       0
     59 TotalTechnologyAnnualActivity(BB,gas_plant,2016)                                2.1101                       0
     60 TotalTechnologyModelPeriodActivity(BB,gas_import)                             2.342422                       0
     61 TotalTechnologyModelPeriodActivity(BB,gas_plant)                                2.1101                       0
     64 ProductionByTechnologyAnnual(BB,gas_import,natural_gas,2016)                  2.342422                       0
     65 ProductionByTechnologyAnnual(BB,gas_plant,electricity,2016)                     2.1101                       0
     68 UseByTechnologyAnnual(BB,gas_plant,natural_gas,2016)                          2.342422                       0
